"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Component = Component;

/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
function Component() {
  const [count] = require('react').useState(0);

  return count;
}
//# sourceMappingURL=InlineRequire.js.map?foo=bar&param=some_value