'use strict';

if (process.env.NODE_ENV === 'production') {
  module.exports = require('./cjs/react-server-dom-webpack-writer.node.production.min.server.js');
} else {
  module.exports = require('./cjs/react-server-dom-webpack-writer.node.development.server.js');
}
