.List {
  /**
   * The decision to suppress the default (horizontal) scroll bars is intentional.
   * The snapshot selector is small enough that the scrollbar renders it essentially useless.
   * There are two alternative navigation mechanisms (arrow keys and arrow buttons).
   * If necessary, we could also expand the click-and-drag mechanism to scroll.
   */
  overflow-x: hidden !important;
}

.Tooltip {
  margin-top: 2.5rem;
}

.TooltipList {
  list-style: none;
  padding: 0;
  margin: 0;
}

.TooltipListItem {
  display: flex;
}

.TooltipLabel {
  font-weight: bold;
  margin-right: 0.25rem;
}
.TooltipLabel:after {
  content: ':';
}

.TooltipValue {
  flex-grow: 1;
  text-align: end;
}

.DurationsWrapper {
  flex-grow: 1;
}

.DurationsList {
  list-style: none;
  margin: 0 0 0 1rem;
  padding: 0;
}

.DurationsListItem {
  display: flex;
}

.DurationsLabel {
  margin-right: 0.25rem;
}
.DurationsLabel:after {
  content: ':';
}

.DurationsValue {
  flex-grow: 1;
  text-align: end;
}