.Outer {
  user-select: none;
  cursor: pointer;
  border-bottom: 3px solid transparent;
  padding-top: 4px;
  padding-bottom: 1px;
  display: flex;
  align-items: flex-end;
}
.Outer:hover {
  background-color: var(--color-background);
}

.Inner,
.InnerSelected {
  width: 100%;
  min-height: 2px;
  background-color: var(--color-commit-did-not-render-fill);
  color: var(--color-commit-did-not-render-fill-text);
}

.InnerSelected {
  background-color: var(--color-button-active);
}