:root {
  /* Font smoothing */
  --font-smoothing: auto;

  /* GitHub.com system fonts */
  --font-family-monospace: 'SFMono-Regular', Consolas, 'Liberation Mono', Menlo,
    Courier, monospace;
  --font-family-sans: -apple-system, BlinkMacSystemFont, Segoe UI, Helvetica,
    Arial, sans-serif, Apple Color Emoji, Segoe UI Emoji, Segoe UI Symbol;
}
