body {
  font-family: system-ui, sans-serif;
}

* {
  box-sizing: border-box;
}

nav {
  padding: 20px;
}

.sidebar {
  padding: 10px;
  height: 500px;
  float: left;
  width: 30%;
}

.post {
  padding: 20px;
  float: left;
  width: 60%;
}

h1, h2 {
  padding: 0;
}

ul, li {
  margin: 0;
}

.post p {
  font-size: larger;
  font-family: Georgia, serif;
}

.comments {
  margin-top: 40px;
}

.comment {
  border: 2px solid #aaa;
  border-radius: 4px;
  padding: 20px;
}

/* https://codepen.io/mandelid/pen/vwKoe */
.spinner {
  display: inline-block;
  transition: opacity linear 0.1s;
  width: 20px;
  height: 20px;
  border: 3px solid rgba(80, 80, 80, 0.5);
  border-radius: 50%;
  border-top-color: #fff;
  animation: spin 1s ease-in-out infinite;
  opacity: 0;
}
.spinner--active {
  opacity: 1;
}

@keyframes spin {
  to {
    transform: rotate(360deg);
  }
}
@keyframes spin {
  to {
    transform: rotate(360deg);
  }
}