.Updaters {
  margin: 0 0 0.5rem;
}

.NoUpdaters,
.Updater,
.UnmountedUpdater {
  display: block;
  width: 100%;
  text-align: left;
  background: none;
  border: none;
  padding: 0.25rem 0.5rem;
  color: var(--color-text);
}
.Updater:focus,
.Updater:hover {
  outline: none;
  background-color: var(--color-background-hover);
}

.NoUpdaters,
.UnmountedUpdater {
  color: var(--color-dim);
}
