.Content {
  width: 100%;
  position: relative;
  flex: 1 1 auto;
  display: flex;
  flex-direction: column;
  align-items: center;
  justify-content: center;
}

.Paragraph {
  text-align: center;
}

.ErrorMessage {
  margin: 0.5rem 0;
  color: var(--color-dim);
  font-family: var(--font-family-monospace);
  font-size: var(--font-size-monospace-normal);
}

.Row {
  display: flex;
  flex-direction: row;
  align-items: center;
  justify-content: center;
  flex-flow: wrap;
}

.EmptyStateContainer {
  text-align: center;
}

.Header {
  font-size: var(--font-size-sans-large);
  margin-bottom: 0.5rem;
}

.WelcomeInstructionsList {
}

.WelcomeInstructionsListItem {
  display: flex;
  align-items: center;
  line-height: 1.5rem;
  counter-increment: li;
}

.WelcomeInstructionsListItem::before {
  content: counter(li);
  margin-right: 0.5rem;
}

.WelcomeInstructionsListItemLink {
  color: var(--color-link);
  margin-left: 0.25rem;
  margin-right: 0.25rem;
}

.ImportButtonLabel {
  margin-left: 0.25rem;
}