/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow
 */

declare var __EXTENSION__: boolean;
declare var __TEST__: boolean;

declare var DARK_MODE_DIMMED_WARNING_COLOR: string;
declare var DARK_MODE_DIMMED_ERROR_COLOR: string;
declare var DARK_MODE_DIMMED_LOG_COLOR: string;
declare var LIGHT_MODE_DIMMED_WARNING_COLOR: string;
declare var LIGHT_MODE_DIMMED_ERROR_COLOR: string;
declare var LIGHT_MODE_DIMMED_LOG_COLOR: string;
